# stdlib
from typing import Any, Sequence, Tuple, Optional, Union

# third party
import numpy as np

# pygram11
from pygram11._backend import (
    _f1d,
    _v1d,
    _f1dw,
    _v1dw,
    _f1dmw,
    _v1dmw,
    _f2d,
    _f2dw,
    _v2d,
    _v2dw,
)
from ._helpers import limits_1d, limits_2d, likely_uniform_bins


def _densify_fixed_counts(counts: np.ndarray, width: float) -> np.ndarray:
    """Convert fixed width histogram to unity integral over PDF."""
    return np.array(counts / (width * counts.sum()), dtype=np.float64)


def _densify_fixed_weighted_counts(
    raw: Tuple[np.ndarray, np.ndarray], width: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert fixed width weighted histogram to unity integral over PDF."""
    counts = raw[0]
    integral = counts.sum()
    res0 = _densify_fixed_counts(counts, width)
    variances = raw[1]
    f1 = 1.0 / ((width * integral) ** 2)
    f2 = counts / integral
    res1 = f1 * (variances + (f2 * f2 * variances.sum()))
    return res0, np.sqrt(res1)


def _densify_variable_counts(
    counts: np.ndarray,
    edges: np.ndarray,
) -> np.ndarray:
    """Convert variable width histogram to unity integral over PDF."""
    widths = edges[1:] - edges[:-1]
    integral = float(np.sum(counts))
    return np.array(counts / widths / integral, dtype=np.float64)


def _densify_variable_weighted_counts(
    raw: Tuple[np.ndarray, np.ndarray], edges: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Convert variable width histogram to unity integral over PDF."""
    counts = raw[0]
    variances = raw[1]
    integral = counts.sum()
    widths = edges[1:] - edges[:-1]
    res0 = _densify_variable_counts(counts, edges)
    f1 = 1.0 / ((widths * integral) ** 2)
    f2 = counts / integral
    res1 = f1 * (variances + (f2 * f2 * variances.sum()))
    return res0, res1


def bin_edges(bins: int, range: Tuple[float, float]) -> np.ndarray:
    """Construct bin edges given number of bins and axis limits.

    Parameters
    ----------
    bins : int
        Total number of bins.
    range : (float, float)
        Minimum and maximum of the histogram axis.

    Returns
    -------
    numpy.ndarray
        Edges defined by the number of bins and axis limits.

    Examples
    --------
    >>> bin_edges(bins=8, range=(-2, 2))
    array([-2. , -1.5, -1. , -0.5,  0. ,  0.5,  1. ,  1.5,  2. ])

    """
    return np.linspace(range[0], range[1], bins + 1)


def bin_centers(
    bins: Union[int, Sequence[float], np.ndarray],
    range: Optional[Tuple[float, float]] = None,
) -> np.ndarray:
    """Construct array of center values for each bin.

    Parameters
    ----------
    bins : int or array_like
        Number of bins or bin edges array.
    range : (float, float), optional
        The minimum and maximum of the histogram axis.

    Returns
    -------
    numpy.ndarray
        Array of bin centers.

    Raises
    ------
    ValueError
        If `bins` is an integer and range is undefined (``None``).

    Examples
    --------
    The centers given the number of bins and max/min:

    >>> bin_centers(10, range=(-3, 3))
    array([-2.7, -2.1, -1.5, -0.9, -0.3,  0.3,  0.9,  1.5,  2.1,  2.7])

    Or given bin edges:

    >>> bin_centers([0, 1, 2, 3, 4])
    array([0.5, 1.5, 2.5, 3.5])

    """
    if isinstance(bins, int):
        if range is None:
            raise ValueError("Integer bins requires defining range")
        bins = bin_edges(bins, range=range)
    b = np.asarray(bins)
    return 0.5 * (b[1:] + b[:-1])


def fix1d(
    x: np.ndarray,
    bins: int = 10,
    range: Optional[Tuple[float, float]] = None,
    weights: Optional[np.ndarray] = None,
    density: bool = False,
    flow: bool = False,
    cons_var: bool = False,
) -> Tuple[np.ndarray, Union[np.ndarray, None]]:
    r"""Histogram data with fixed (uniform) bin widths.

    Parameters
    ----------
    x : numpy.ndarray
        Data to histogram.
    bins : int
        The number of bins.
    range : (float, float), optional
        The minimum and maximum of the histogram axis. If ``None``,
        min and max of `x` will be used.
    weights : numpy.ndarray, optional
        The weights for each element of `x`. If weights are absent,
        the second return type will be ``None``.
    density : bool
        Normalize histogram counts as value of PDF such that the
        integral over the range is unity.
    flow : bool
        Include under/overflow in the first/last bins.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If `x` and `weights` have incompatible shapes.
    TypeError
        If `x` or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The resulting histogram bin counts.
    :py:obj:`numpy.ndarray`, optional
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. The return is ``None`` if weights are not used. If
        `cons_var` is ``True``, the variances are returned.

    Examples
    --------
    A histogram of `x` with 20 bins between 0 and 100:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.uniform(0, 100, size=(100,))
    >>> h, __ = fix1d(x, bins=20, range=(0, 100))

    When weights are absent the second return is ``None``. The same
    data, now histogrammed with weights and over/underflow included:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.uniform(0, 100, size=(100,))
    >>> w = rng.uniform(0.1, 0.9, x.shape[0])
    >>> h, stderr = fix1d(x, bins=20, range=(0, 100), weights=w, flow=True)

    """
    xmin, xmax = limits_1d(x, range)

    if weights is None:
        result = _f1d(x, bins, xmin, xmax, flow)
        if density:
            width = (xmax - xmin) / bins
            result = _densify_fixed_counts(result, width)
        return result, None

    if np.shape(x) != np.shape(weights):
        raise ValueError("x and weights must have the same shape")

    result = _f1dw(x, weights, int(bins), xmin, xmax, flow)
    if density:
        width = (xmax - xmin) / bins
        result = _densify_fixed_weighted_counts(result, width)
    counts, variances = result

    if cons_var:
        return counts, variances
    return counts, np.sqrt(variances)


def fix1dmw(
    x: np.ndarray,
    weights: np.ndarray,
    bins: int = 10,
    range: Optional[Tuple[float, float]] = None,
    flow: bool = False,
    cons_var: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Histogram data with multiple weight variations and fixed width bins.

    The weights array must have a total number of rows equal to the
    length of the input data. The number of columns in the weights
    array is equal to the number of weight variations. (The weights
    array must be an `M x N` matrix where `M` is the length of `x` and
    `N` is the number of weight variations).

    Parameters
    ----------
    x : numpy.ndarray
        Data to histogram.
    weights : numpy.ndarray
        The weight variations for the elements of `x`, first
        dimension is the length of `x`, second dimension is the
        number of weights variations.
    bins : int
        The number of bins.
    range : (float, float), optional
        The minimum and maximum of the histogram axis. If ``None``,
        min and max of `x` will be used.
    flow : bool
        Include under/overflow in the first/last bins.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If `x` and `weights` have incompatible shapes (if
        ``x.shape[0] != weights.shape[0]``).
    ValueError
        If `weights` is not a two dimensional array.
    TypeError
        If `x` or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. If `cons_var` is ``True``, the variances are
        returned.

    Examples
    --------
    Multiple histograms of `x` using 20 different weight variations:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.standard_normal(10000)
    >>> twenty_weights = np.abs(rng.standard_normal((x.shape[0], 20)))
    >>> h, err = fix1dmw(x, twenty_weights, bins=50, range=(-3, 3))

    `h` and `err` are now shape ``(50, 20)``. Each column represents
    the histogram of the data using its respective weight.

    """
    if len(np.shape(weights)) != 2:
        raise ValueError("weights must be a two dimensional array.")

    if np.shape(x)[0] != np.shape(weights)[0]:
        raise ValueError("x and weights have incompatible shapes.")

    xmin, xmax = limits_1d(x, range)
    counts, variances = _f1dmw(x, weights, int(bins), xmin, xmax, flow)
    if cons_var:
        return counts, variances
    return counts, np.sqrt(variances)


def var1d(
    x: np.ndarray,
    bins: np.ndarray,
    weights: Optional[np.ndarray] = None,
    density: bool = False,
    flow: bool = False,
    cons_var: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    r"""Histogram data with variable bin widths.

    Parameters
    ----------
    x : numpy.ndarray
        Data to histogram
    bins : numpy.ndarray
        Bin edges
    weights : numpy.ndarray, optional
        The weights for each element of `x`. If weights are absent,
        the second return type will be ``None``.
    density : bool
        Normalize histogram counts as value of PDF such that the
        integral over the range is unity.
    flow : bool
        Include under/overflow in the first/last bins.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If the array of bin edges is not monotonically increasing.
    ValueError
        If `x` and `weights` have incompatible shapes.
    TypeError
        If `x` or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`, optional
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. If `cons_var` is True, the variances are returned.
        The return is ``None`` if weights are not used.

    Examples
    --------
    A simple histogram with variable width bins:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.standard_normal(1000)
    >>> edges = np.array([-3.0, -2.5, -1.5, -0.25, 0.25, 2.0, 3.0])
    >>> h, __ = var1d(x, edges)

    """
    if not np.all(bins[1:] >= bins[:-1]):
        raise ValueError("bins sequence must monotonically increase")

    if likely_uniform_bins(bins):
        nbins = np.shape(bins)[0] - 1
        return fix1d(
            x,
            bins=nbins,
            weights=weights,
            range=(bins[0], bins[-1]),
            flow=flow,
            density=density,
        )

    bins = np.array(bins, dtype=np.float64, copy=False)
    if weights is None:
        result = _v1d(x, bins, flow)
        if density:
            result = _densify_variable_counts(result, bins)
        return result, None

    if np.shape(x) != np.shape(weights):
        raise ValueError("x and weights have incompatible shapes.")

    result = _v1dw(x, weights, bins, flow)
    if density:
        result = _densify_variable_weighted_counts(result, bins)
    counts, variances = result
    if cons_var:
        return counts, variances
    return counts, np.sqrt(variances)


def var1dmw(
    x: np.ndarray,
    weights: np.ndarray,
    bins: np.ndarray,
    flow: bool = False,
    cons_var: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    r"""Histogram data with multiple weight variations and variable width bins.

    The weights array must have a total number of rows equal to the
    length of the input data. The number of columns in the weights
    array is equal to the number of weight variations. (The weights
    array must be an `M x N` matrix where `M` is the length of `x` and
    `N` is the number of weight variations).

    Parameters
    ----------
    x : numpy.ndarray
        Data to histogram.
    weights : numpy.ndarray
        Weight variations for the elements of `x`, first dimension
        is the shape of `x`, second dimension is the number of weights.
    bins : numpy.ndarray
        Bin edges.
    flow : bool
        Include under/overflow in the first/last bins.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If the array of bin edges is not monotonically increasing.
    ValueError
        If `x` and `weights` have incompatible shapes.
    ValueError
        If `weights` is not a two dimensional array.
    TypeError
        If `x` or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. If `cons_var` is ``True``, the variances are
        returned.

    Examples
    --------
    Using three different weight variations:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.standard_normal(10000)
    >>> weights = np.abs(rng.standard_normal((x.shape[0], 3)))
    >>> edges = np.array([-3.0, -2.5, -1.5, -0.25, 0.25, 2.0, 3.0])
    >>> h, err = var1dmw(x, weights, edges)
    >>> h.shape
    (6, 3)
    >>> err.shape
    (6, 3)

    """
    if len(np.shape(weights)) != 2:
        raise ValueError("weights must be a two dimensional array.")
    if np.shape(x)[0] != np.shape(weights)[0]:
        raise ValueError("x and weights have incompatible shapes.")
    if not np.all(bins[1:] >= bins[:-1]):
        raise ValueError("bins sequence must monotonically increase.")

    if likely_uniform_bins(bins):
        return fix1dmw(
            x,
            weights,
            bins=(len(bins) - 1),
            range=(bins[0], bins[-1]),
            flow=flow,
        )

    counts, variances = _v1dmw(x, weights, bins, flow)
    if cons_var:
        return counts, variances
    return counts, np.sqrt(variances)


def fix2d(
    x: np.ndarray,
    y: np.ndarray,
    bins: Union[int, Tuple[int, int]] = 10,
    range: Optional[Sequence[Tuple[float, float]]] = None,
    weights: Optional[np.ndarray] = None,
    flow: bool = False,
    cons_var: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    r"""Histogram two dimensional data with fixed (uniform) binning.

    The two input arrays (`x` and `y`) must be the same length
    (shape).

    Parameters
    ----------
    x : numpy.ndarray
        First entries in data pairs to histogram.
    y : numpy.ndarray
        Second entries in data pairs to histogram.
    bins : int or (int, int)
        If int, both dimensions will have that many bins;
        if tuple, the number of bins for each dimension
    range : Sequence[Tuple[float, float]], optional
        Axis limits in the form ``[(xmin, xmax), (ymin, ymax)]``. If
        ``None`` the input data min and max will be used.
    weights : array_like, optional
        The weights for data element. If weights are absent, the
        second return type will be ``None``.
    flow : bool
        Include over/underflow.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If `x` and `y` have incompatible shapes.
    ValueError
        If the shape of `weights` is incompatible with `x` and `y`
    TypeError
        If `x`, `y`, or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`, optional
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. If `cons_var` is ``True``, the variances are
        returned.

    Examples
    --------
    A histogram of (`x`, `y`) with 20 bins between 0 and 100 in
    the `x` dimention and 10 bins between 0 and 50 in the `y`
    dimension:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.uniform(0, 100, size=(200,))
    >>> y = rng.uniform(0, 50, size=(200,))
    >>> h, __ = fix2d(x, y, bins=(20, 10), range=((0, 100), (0, 50)))

    The same data, now histogrammed weighted (via `w`):

    >>> w = rng.uniform(0.2, 0.9, size=x.shape)
    >>> h, err = fix2d(x, y, bins=(20, 10), range=((0, 100), (0, 50)), weights=w)

    """
    if np.shape(x) != np.shape(y):
        raise ValueError("x and y must be the same shape.")
    if weights is not None:
        if np.shape(weights) != np.shape(x):
            raise ValueError("data and weights must be the same shape.")

    if isinstance(bins, int):
        nx = ny = bins
    else:
        nx, ny = bins

    xmin, xmax, ymin, ymax = limits_2d(x, y, range)

    if weights is None:
        result = _f2d(x, y, int(nx), xmin, xmax, int(ny), ymin, ymax, flow)
        return result, None

    counts, variances = _f2dw(
        x, y, weights, int(nx), xmin, xmax, int(ny), ymin, ymax, flow
    )

    if cons_var:
        return counts, variances
    return counts, np.sqrt(variances)


def var2d(
    x: np.ndarray,
    y: np.ndarray,
    xbins: np.ndarray,
    ybins: np.ndarray,
    weights: Optional[np.ndarray] = None,
    flow: bool = False,
    cons_var: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    r"""Histogram two dimensional data with variable width binning.

    The two input arrays (`x` and `y`) must be the same length
    (shape).

    Parameters
    ----------
    x : numpy.ndarray
        First entries in data pairs to histogram.
    y : numpy.ndarray
        Second entries in data pairs to histogram.
    xbins : numpy.ndarray
        Bin edges for the `x` dimension.
    ybins : np.ndarray
        Bin edges for the `y` dimension.
    weights : array_like, optional
        The weights for data element. If weights are absent, the
        second return type will be ``None``.
    flow : bool
        Include under/overflow.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If `x` and `y` have different shape.
    ValueError
        If either bin edge definition is not monotonically increasing.
    TypeError
        If `x`, `y`, or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`, optional
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. If `cons_var` is ``True``, the variances are
        returned.

    Examples
    --------
    A histogram of (`x`, `y`) where the edges are defined by a
    :func:`numpy.logspace` in both dimensions:

    >>> x = np.exp(np.random.uniform(0, 1, size=(10000,)))
    >>> y = np.exp(np.random.uniform(0, 1, size=(10000,)))
    >>> bins = np.logspace(0.1, 1.0, 10, endpoint=True)
    >>> h, __ = var2d(x, y, bins, bins)

    """
    if np.shape(x) != np.shape(y):
        raise ValueError("x and y must be the same shape.")
    if not np.all(xbins[1:] >= xbins[:-1]):
        raise ValueError("xbins sequence must monotonically increase.")
    if not np.all(ybins[1:] >= ybins[:-1]):
        raise ValueError("ybins sequence must monotonically increase.")
    if weights is not None:
        weights = np.asarray(weights)
        if np.shape(weights) != np.shape(x):
            raise ValueError("data and weights must be the same shape.")

    if weights is None:
        result = _v2d(x, y, xbins, ybins, flow)
        return result, None

    counts, variances = _v2dw(x, y, weights, xbins, ybins, flow)
    if cons_var:
        return counts, variances
    return counts, np.sqrt(variances)



import numpy as np
from pygram11._helpers import limits_1d, limits_2d, likely_uniform_bins


def test_limits_1d():
    x = np.array([1, 5, 2, 18, 202, 0.5, 0.5])
    assert limits_1d(x) == (0.5, 202)
    assert limits_1d(x, range=(1, 505)) == (1, 505)


def test_limits_2d():
    x = np.array([3, 1, 3, 18, 21, 19])
    y = np.array([5, 9, 3, 25, 12, 88])
    ranges = ((5, 20), (4, 18))
    assert limits_2d(x, y) == (1, 21, 3, 88)
    assert limits_2d(x, y, range=ranges) == (5, 20, 4, 18)


def test_likely_uniform_bins():
    b1 = np.array([1, 2, 3, 4, 5])
    b2 = np.array([1.1, 2, 3, 4])
    assert likely_uniform_bins(b1)
    assert not likely_uniform_bins(b2)



import matplotlib.pyplot as plt
import matplotlib as mpl

mpl.use("Agg")

import json
import sys
from timeit import timeit, repeat
import pygram11 as pg
import numpy as np

SETUP_F_1D = """
import numpy as np
import boost_histogram as bh
from numpy import histogram as np_hist
from pygram11 import histogram as pg_hist
from fast_histogram import histogram1d as fh_hist
rng = np.random.default_rng(123)
x = rng.standard_normal({size})
w = rng.uniform(0.4, 0.8, {size})
bh_hist = bh.Histogram(bh.axis.Regular(25, -3, 3), storage=bh.storage.Weight())
"""

NP_F_1D = "np_hist(x, bins=25, range=(-3, 3), weights=w)"
PG_F_1D = "pg_hist(x, bins=25, range=(-3, 3), weights=w)"
FH_F_1D = "fh_hist(x, bins=25, range=(-3, 3), weights=w)"
BH_F_1D = "bh_hist.fill(x, weight=w)"

################################################################

SETUP_V_1D = """
import numpy as np
import boost_histogram as bh
from numpy import histogram as np_hist
from pygram11 import histogram as pg_hist
rng = np.random.default_rng(123)
edges = np.unique(np.round(sorted(rng.uniform(-3, 3, 26)), 2))
x = rng.standard_normal({size})
w = rng.uniform(0.4, 0.8, {size})
bh_hist = bh.Histogram(bh.axis.Variable(edges), storage=bh.storage.Weight())
"""

NP_V_1D = "np_hist(x, bins=edges, weights=w)"
PG_V_1D = "pg_hist(x, bins=edges, weights=w)"
BH_V_1D = "bh_hist.fill(x, weight=w)"

################################################################

SETUP_F_2D = """
import numpy as np
import boost_histogram as bh
from numpy import histogram2d as np_hist
from pygram11 import histogram2d as pg_hist
from fast_histogram import histogram2d as fh_hist
rng = np.random.default_rng(123)
x = rng.standard_normal({size})
y = rng.standard_normal({size})
w = rng.uniform(0.4, 0.8, {size})
bh_hist = bh.Histogram(bh.axis.Regular(25, -3, 3),
                       bh.axis.Regular(25, -3, 3),
                       storage=bh.storage.Weight())
"""

NP_F_2D = "np_hist(x, y, bins=[25, 25], range=[[-3, 3], [-3, 3]], weights=w)"
PG_F_2D = "pg_hist(x, y, bins=[25, 25], range=[[-3, 3], [-3, 3]], weights=w)"
FH_F_2D = "fh_hist(x, y, bins=[25, 25], range=[[-3, 3], [-3, 3]], weights=w)"
BH_F_2D = "bh_hist.fill(x, y, weight=w)"

################################################################

SETUP_V_2D = """
import numpy as np
import boost_histogram as bh
from numpy import histogram as np_hist
from pygram11 import histogram as pg_hist
edges = np.unique(np.round(sorted(rng.uniform(-3, 3, 26)), 2))
rng = np.random.default_rng(123)
x = rng.standard_normal({size})
y = rng.standard_normal({size})
w = rng.uniform(0.4, 0.8, {size})
bh_hist = bh.Histogram(bh.axis.Variable(edges),
                       bh.axis.Variable(edges),
                       storage=bh.storage.Weight())
"""

NP_V_2D = "np_hist(x, bins=edges, weights=w)"
PG_V_2D = "pg_hist(x, bins=edges, weights=w)"
BH_V_2D = "bh_hist.fill(x, y, weight=w)"

################################################################

expos1d = [2, 3, 4, 5, 6, 7, 8]
expos2d = [2, 3, 4, 5, 6, 7, 8]


def time_stats(stmt=None, setup=None):
    time_single = timeit(stmt=stmt, setup=setup, number=1)
    number = max(3, int(1.0 / time_single))
    print(
        " -> estimated time to complete test: {0:.1f}s".format(
            time_single * 10 * number
        )
    )
    times = repeat(stmt=stmt, setup=setup, repeat=10, number=number)
    return np.min(times) / number, np.mean(times) / number, np.median(times) / number


def run_4(expos, stmt_pg, stmt_np, stmt_bh, stmt_fh, setup, outname):
    pg_means = []
    np_means = []
    fh_means = []
    bh_means = []
    for exponent in expos:
        size = int(10 ** exponent)
        print(f"Running 10^{exponent}")
        pg_min, pg_mean, pg_median = time_stats(stmt=stmt_pg, setup=setup.format(size=size))
        np_min, np_mean, np_median = time_stats(stmt=stmt_np, setup=setup.format(size=size))
        fh_min, fh_mean, fh_median = time_stats(stmt=stmt_fh, setup=setup.format(size=size))
        bh_min, bh_mean, bh_median = time_stats(stmt=stmt_bh, setup=setup.format(size=size))
        np_means.append(np_mean)
        pg_means.append(pg_mean)
        fh_means.append(fh_mean)
        bh_means.append(bh_mean)
        print(exponent, np_min, np_mean, np_median)
        print(exponent, pg_min, pg_mean, pg_median)
        print(exponent, fh_min, fh_mean, fh_median)
        print(exponent, bh_min, bh_mean, bh_median)
    res = {
        "pg_means": pg_means,
        "np_means": np_means,
        "fh_means": fh_means,
        "bh_means": bh_means,
    }
    with open(outname, "w") as f:
        json.dump(res, f, indent=4)


def run_3(expos, stmt_pg, stmt_np, stmt_bh, setup, outname):
    pg_means = []
    np_means = []
    bh_means = []
    for exponent in expos:
        size = int(10 ** exponent)
        print(f"Running 10^{exponent}")
        pg_min, pg_mean, pg_median = time_stats(stmt=stmt_pg, setup=setup.format(size=size))
        np_min, np_mean, np_median = time_stats(stmt=stmt_np, setup=setup.format(size=size))
        bh_min, bh_mean, bh_median = time_stats(stmt=stmt_bh, setup=setup.format(size=size))
        np_means.append(np_mean)
        pg_means.append(pg_mean)
        bh_means.append(bh_mean)
        print(exponent, np_min, np_mean, np_median)
        print(exponent, pg_min, pg_mean, pg_median)
        print(exponent, bh_min, bh_mean, bh_median)
    res = {
        "pg_means": pg_means,
        "np_means": np_means,
        "bh_means": bh_means,
    }
    with open(outname, "w") as f:
        json.dump(res, f, indent=3)


def plots_4(inname, title, logscale=True):
    with open(inname, "r") as f:
        res = json.load(f)
    np_means = np.array(res["np_means"])
    fh_means = np.array(res["fh_means"])
    bh_means = np.array(res["bh_means"])
    pg_means = np.array(res["pg_means"])
    fig, ax = plt.subplots()
    ax.plot(expos1d, np_means / pg_means, label="over numpy")
    ax.plot(expos1d, fh_means / pg_means, label="over fast_histogram")
    ax.plot(expos1d, bh_means / pg_means, label="over boost_histogram")
    ax.grid()
    if logscale:
        ax.set_yscale("log")
    ax.set_title(title)
    ax.set_ylabel("Speedup")
    ax.set_xlabel("Array Size")
    ax.set_xticks(expos1d)
    ax.set_xticklabels([f"$10^{n}$" for n in expos1d])
    ax.legend(loc="best")
    fig.savefig(f"bm_{inname}.png")


def plots_3(inname, title, logscale=True):
    with open(inname, "r") as f:
        res = json.load(f)
    np_means = np.array(res["np_means"])
    bh_means = np.array(res["bh_means"])
    pg_means = np.array(res["pg_means"])
    fig, ax = plt.subplots()
    ax.plot(expos1d, np_means / pg_means, label="over numpy")
    ax.plot(expos1d, bh_means / pg_means, label="over boost_histogram")
    ax.grid()
    if logscale:
        ax.set_yscale("log")
    ax.set_title(title)
    ax.set_ylabel("Speedup")
    ax.set_xlabel("Array Size")
    ax.set_xticks(expos1d)
    ax.set_xticklabels([f"$10^{n}$" for n in expos1d])
    ax.legend(loc="best")
    fig.savefig(f"bm_{inname}.png")


if len(sys.argv) < 2:
    print("what do you want to do")
    exit(1)
if "fixed1d" in sys.argv:
    run_4(expos1d, PG_F_1D, NP_F_1D, BH_F_1D, FH_F_1D, SETUP_F_1D, "fixed1d.json")
if "var1d" in sys.argv:
    run_3(expos1d, PG_V_1D, NP_V_1D, BH_V_1D, SETUP_V_1D, "var1d.json")
if "fixed2d" in sys.argv:
    run_4(expos2d, PG_F_2D, NP_F_2D, BH_F_2D, FH_F_2D, SETUP_F_2D, "fixed2d.json")
if "var2d" in sys.argv:
    run_3(expos2d, PG_V_2D, NP_V_2D, BH_V_2D, SETUP_V_2D, "var2d.json")
if "plotvar1d" in sys.argv:
    plots_3("var1d.json", "Variable Width 1D Histograms")
if "plotfixed1d" in sys.argv:
    plots_4("fixed1d.json", "Fixed Width 1D Histograms")
if "plotfixed2d" in sys.argv:
    plots_4("fixed2d.json", "Fixed Width 2D Histograms")
if "plotvar2d" in sys.argv:
    plots_3("var2d.json", "Variable Width 2D Histograms")



# stdlib
from typing import Optional, Sequence, Tuple

# third party
import numpy as np


def limits_1d(
    x: np.ndarray, range: Optional[Tuple[float, float]] = None
) -> Tuple[float, float]:
    """Get bin limits given an optional range and 1D data."""
    if range is None:
        return (float(np.amin(x)), float(np.amax(x)))
    return float(range[0]), float(range[1])


def limits_2d(
    x: np.ndarray,
    y: np.ndarray,
    range: Optional[Sequence[Tuple[float, float]]] = None,
) -> Tuple[float, float, float, float]:
    """Get bin limits given an optional range and 2D data."""
    if range is None:
        return (
            float(np.amin(x)),
            float(np.amax(x)),
            float(np.amin(y)),
            float(np.amax(y)),
        )
    else:
        return (
            float(range[0][0]),
            float(range[0][1]),
            float(range[1][0]),
            float(range[1][1]),
        )


def likely_uniform_bins(edges: np.ndarray) -> bool:
    """Test if bin edges describe a set of fixed width bins."""
    diffs = np.ediff1d(edges)
    return bool(np.all(np.isclose(diffs, diffs[0])))



import os
import multiprocessing
import sys
import pytest
import pygram11
import pygram11.config
import pygram11 as pg


TKEYS = list(pygram11.config.threshold_keys())


def test_omp_get_max_threads():
    nthreads = os.getenv("OMP_NUM_THREADS")
    if nthreads is None:
        nthreads = multiprocessing.cpu_count()
    assert int(nthreads) == pg.omp_get_max_threads()


class TestThresholds:
    @pytest.mark.parametrize("k", TKEYS)
    def test_contextmanagers_keys(self, k):
        pygram11.default_omp()
        with pygram11.omp_disabled(key=k):
            assert pygram11.config.get(k) == sys.maxsize
            for tk in TKEYS:
                if tk != k:
                    assert pygram11.config.get(tk) != sys.maxsize
                    assert pygram11.config.get(tk) != 0

        for tk in TKEYS:
            assert pygram11.config.get(tk) != sys.maxsize
            assert pygram11.config.get(tk) != 0

        with pygram11.omp_forced(key=k):
            assert pygram11.config.get(k) == 0
            for tk in TKEYS:
                if tk != k:
                    assert pygram11.config.get(tk) != sys.maxsize
                    assert pygram11.config.get(tk) != 0

    def test_contextmanagers_blanket(self):
        pygram11.default_omp()
        with pygram11.omp_disabled():
            for tk in TKEYS:
                assert pygram11.config.get(tk) == sys.maxsize

        for tk in TKEYS:
            assert pygram11.config.get(tk) != sys.maxsize
            assert pygram11.config.get(tk) != 0

        with pygram11.omp_forced():
            for tk in TKEYS:
                assert pygram11.config.get(tk) == 0

        for tk in TKEYS:
            assert pygram11.config.get(tk) != sys.maxsize
            assert pygram11.config.get(tk) != 0

    @pytest.mark.parametrize("k", TKEYS)
    def test_decorators(self, k):
        @pygram11.with_omp(key=k)
        def f():
            return pygram11.config.get(k)

        assert f() == 0

        @pygram11.without_omp(key=k)
        def g():
            return pygram11.config.get(k)

        assert g() == sys.maxsize

        @pygram11.with_omp
        def h():
            return [pygram11.config.get(tk) for tk in TKEYS]

        assert h() == [0 for tk in TKEYS]

        @pygram11.without_omp
        def j():
            return [pygram11.config.get(tk) for tk in TKEYS]

        assert j() == [sys.maxsize for tk in TKEYS]



# MIT License
#
# Copyright (c) 2021 Douglas Davis
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import functools
import contextlib
import sys
from typing import Iterator, Optional
import pygram11.config
from pygram11._backend import _omp_get_max_threads


def omp_get_max_threads() -> int:
    """Get the number of threads available to OpenMP.

    This returns the result of calling the OpenMP C API function `of
    the same name
    <https://www.openmp.org/spec-html/5.0/openmpsu112.html>`_.

    Returns
    -------
    int
        the maximum number of available threads

    """
    return _omp_get_max_threads()


def default_omp() -> None:
    """Set OpenMP acceleration thresholds to the default values."""
    pygram11.config.set("thresholds.fix1d", 10_000)
    pygram11.config.set("thresholds.fix1dmw", 10_000)
    pygram11.config.set("thresholds.fix2d", 10_000)
    pygram11.config.set("thresholds.var1d", 5_000)
    pygram11.config.set("thresholds.var1dmw", 5_000)
    pygram11.config.set("thresholds.var2d", 5_000)


def disable_omp() -> None:
    """Disable OpenMP acceleration by maximizing all thresholds."""
    for k in pygram11.config.threshold_keys():
        pygram11.config.set(k, sys.maxsize)


def force_omp() -> None:
    """Force OpenMP acceleration by nullifying all thresholds."""
    for k in pygram11.config.threshold_keys():
        pygram11.config.set(k, 0)


def without_omp(*args, **kwargs):
    """Wrap a function to disable OpenMP while it's called.

    If a specific key is defined, only that threshold will be modified
    to turn OpenMP off.

    The settings of the pygram11 OpenMP threshold configurations will
    be restored to their previous values at the end of the function
    that is being wrapped.

    Parameters
    ----------
    key : str, optional
        Specific threshold key to turn off.

    Examples
    --------
    Writing a function with this decorator:

    >>> import numpy as np
    >>> from pygram11 import histogram, without_omp
    >>> @without_omp
    ... def single_threaded_histogram():
    ...     data = np.random.standard_normal(size=(1000,))
    ...     return pygram11.histogram(data, bins=10, range=(-5, 5), flow=True)

    Defining a specific `key`:

    >>> import pygram11.config
    >>> previous = pygram11.config.get("thresholds.var1d")
    >>> @without_omp(key="thresholds.var1d")
    ... def single_threaded_histogram2():
    ...     print(f"in function threshold: {pygram11.config.get('thresholds.var1d')}")
    ...     data = np.random.standard_normal(size=(1000,))
    ...     return pygram11.histogram(data, bins=[-2, -1, 1.5, 3.2])
    >>> result = single_threaded_histogram2()
    in function threshold: 9223372036854775807
    >>> previous
    5000
    >>> previous == pygram11.config.get("thresholds.var1d")
    True
    >>> result[0].shape
    (3,)

    """
    func = None
    if len(args) == 1 and callable(args[0]):
        func = args[0]
    if func:
        key = None
    if not func:
        key = kwargs.get("key")

    def cable(func):
        @functools.wraps(func)
        def decorator(*args, **kwargs):
            with omp_disabled(key=key):
                res = func(*args, **kwargs)
            return res

        return decorator

    return cable(func) if func else cable


def with_omp(*args, **kwargs):
    """Wrap a function to always enable OpenMP while it's called.

    If a specific key is defined, only that threshold will be modified
    to turn OpenMP on.

    The settings of the pygram11 OpenMP threshold configurations will
    be restored to their previous values at the end of the function
    that is being wrapped.

    Parameters
    ----------
    key : str, optional
        Specific threshold key to turn on.

    Examples
    --------
    Writing a function with this decorator:

    >>> import numpy as np
    >>> from pygram11 import histogram, with_omp
    >>> @with_omp
    ... def multi_threaded_histogram():
    ...     data = np.random.standard_normal(size=(1000,))
    ...     return pygram11.histogram(data, bins=10, range=(-5, 5), flow=True)

    Defining a specific `key`:

    >>> import pygram11.config
    >>> previous = pygram11.config.get("thresholds.var1d")
    >>> @with_omp(key="thresholds.var1d")
    ... def multi_threaded_histogram2():
    ...     print(f"in function threshold: {pygram11.config.get('thresholds.var1d')}")
    ...     data = np.random.standard_normal(size=(1000,))
    ...     return pygram11.histogram(data, bins=[-2, -1, 1.5, 3.2])
    >>> result = multi_threaded_histogram2()
    in function threshold: 0
    >>> previous
    5000
    >>> previous == pygram11.config.get("thresholds.var1d")
    True
    >>> result[0].shape
    (3,)

    """
    func = None
    if len(args) == 1 and callable(args[0]):
        func = args[0]
    if func:
        key = None
    if not func:
        key = kwargs.get("key")

    def cable(func):
        @functools.wraps(func)
        def decorator(*args, **kwargs):
            with omp_forced(key=key):
                res = func(*args, **kwargs)
            return res

        return decorator

    return cable(func) if func else cable


@contextlib.contextmanager
def omp_disabled(*, key: Optional[str] = None) -> Iterator[None]:
    """Context manager to disable OpenMP.

    Parameters
    ----------
    key : str, optional
        Specific threshold key to turn off.

    Examples
    --------
    Using a specific key:

    >>> import pygram11
    >>> import numpy as np
    >>> with pygram11.omp_disabled(key="thresholds.var1d"):
    ...     data = np.random.standard_normal(size=(200,))
    ...     result = pygram11.histogram(data, bins=[-2, -1, 1.5, 3.2])
    >>> result[0].shape
    (3,)

    Disable all thresholds:

    >>> import pygram11
    >>> import numpy as np
    >>> with pygram11.omp_disabled():
    ...     data = np.random.standard_normal(size=(200,))
    ...     result = pygram11.histogram(data, bins=12, range=(-3, 3))
    >>> result[0].shape
    (12,)

    """
    if key is not None:
        try:
            prev = pygram11.config.get(key)
            pygram11.config.set(key, sys.maxsize)
            yield
        finally:
            pygram11.config.set(key, prev)

    else:
        previous = {k: pygram11.config.get(k) for k in pygram11.config.threshold_keys()}
        try:
            disable_omp()
            yield
        finally:
            for k, v in previous.items():
                pygram11.config.set(k, v)


@contextlib.contextmanager
def omp_forced(*, key: Optional[str] = None) -> Iterator[None]:
    """Context manager to force enable OpenMP.

    Parameters
    ----------
    key : str, optional
        Specific threshold key to turn on.

    Examples
    --------
    Using a specific key:

    >>> import pygram11
    >>> import numpy as np
    >>> with pygram11.omp_forced(key="thresholds.var1d"):
    ...     data = np.random.standard_normal(size=(200,))
    ...     result = pygram11.histogram(data, bins=[-2, -1, 1.5, 3.2])
    >>> result[0].shape
    (3,)

    Enable all thresholds:

    >>> import pygram11
    >>> import numpy as np
    >>> with pygram11.omp_forced():
    ...     data = np.random.standard_normal(size=(200,))
    ...     result = pygram11.histogram(data, bins=10, range=(-3, 3))
    >>> result[0].shape
    (10,)

    """
    if key is not None:
        try:
            prev = pygram11.config.get(key)
            pygram11.config.set(key, 0)
            yield
        finally:
            pygram11.config.set(key, prev)
    else:
        previous = {k: pygram11.config.get(k) for k in pygram11.config.threshold_keys()}
        try:
            force_omp()
            yield
        finally:
            for k, v in previous.items():
                pygram11.config.set(k, v)



"""Simple and fast histogramming in Python.

MIT License

Copyright (c) 2021 Douglas Davis

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

from .version import version as __version__  # noqa

from ._hist import (  # noqa
    bin_centers,
    bin_edges,
    fix1d,
    fix1dmw,
    var1d,
    var1dmw,
    fix2d,
    var2d,
)

from ._numpy import histogram, histogram2d  # noqa

from ._misc import (  # noqa
    omp_get_max_threads,
    force_omp,
    disable_omp,
    default_omp,
    omp_disabled,
    omp_forced,
    without_omp,
    with_omp,
)


version_info = tuple(__version__.split("."))



import numpy as np
from ._hist import fix1d, fix1dmw, fix2d, var1d, var1dmw, var2d


def histogram(
    x,
    bins=10,
    range=None,
    weights=None,
    density=False,
    flow=False,
    cons_var=False,
):
    r"""Histogram data in one dimension.

    Parameters
    ----------
    x : array_like
        Data to histogram.
    bins : int or array_like
        If int: the number of bins; if array_like: the bin edges.
    range : (float, float), optional
        The minimum and maximum of the histogram axis. If ``None``
        with integer `bins`, min and max of `x` will be used. If
        `bins` is an array this is expected to be ``None``.
    weights : array_like, optional
        Weight variations for the elements of `x`. For single weight
        histograms the shape must be the same shape as `x`. For
        multiweight histograms the first dimension is the length of
        `x`, second dimension is the number of weights variations.
    density : bool
        Normalize histogram counts as value of PDF such that the
        integral over the range is unity.
    flow : bool
        Include under/overflow in the first/last bins.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If `bins` defines edges while `range` is also not ``None``.
    ValueError
        If the array of bin edges is not monotonically increasing.
    ValueError
        If `x` and `weights` have incompatible shapes.
    ValueError
        If multiweight histogramming is detected and `weights` is
        not a two dimensional array.
    TypeError
        If `x` or `weights` are unsupported types

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`, optional
        The standard error of each bin count, :math:`\sqrt{\sum_i
        w_i^2}`. The return is ``None`` if weights are not used.

    See Also
    --------
    fix1d
        Used for no weight or single weight fixed bin width histograms
    fix1dmw
        Used for multiweight fixed bin width histograms.
    var1d
        Used for no weight or single weight variable bin width
        histograms.
    var1dmw
        Used for multiweight variable bin width histograms.

    Examples
    --------
    A simple fixed width histogram:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.standard_normal(2000)
    >>> h, __ = histogram(x, bins=20, range=(-3, 3))

    And with variable width histograms and weights:

    >>> w = rng.uniform(0.3, 1.1, size=x.shape)
    >>> h, err = histogram(x, bins=[-3, -2, -1.5, 1.5, 3.5], weights=w)

    """

    # make sure x and weight data are NumPy arrays.
    x = np.array(x)
    is_multiweight = False
    if weights is not None:
        weights = np.asarray(weights)
        is_multiweight = np.shape(weights) != np.shape(x)
        if is_multiweight and len(np.shape(weights)) != 2:
            raise ValueError("weight must be a 2D array for multiweight histograms.")

    # fixed bins
    if isinstance(bins, int):
        if is_multiweight:
            return fix1dmw(
                x,
                weights,
                bins=bins,
                range=range,
                flow=flow,
                cons_var=cons_var,
            )
        return fix1d(
            x,
            weights=weights,
            bins=bins,
            range=range,
            density=density,
            flow=flow,
            cons_var=cons_var,
        )

    # variable bins
    else:
        bins = np.asarray(bins)
        if range is not None:
            raise ValueError("range must be None if bins is non-int")
        if is_multiweight:
            return var1dmw(x, weights, bins=bins, flow=flow, cons_var=cons_var)
        return var1d(
            x,
            weights=weights,
            bins=bins,
            density=density,
            flow=flow,
            cons_var=cons_var,
        )


def histogram2d(x, y, bins=10, range=None, weights=None, flow=False, cons_var=False):
    r"""Histogram data in two dimensions.

    This function provides an API very simiar to
    :func:`numpy.histogram2d`. Keep in mind that the returns are
    different.

    Parameters
    ----------
    x: array_like
        Array representing the `x` coordinate of the data to histogram.
    y: array_like
        Array representing the `y` coordinate of the data to histogram.
    bins: int or array_like or [int, int] or [array, array], optional
        The bin specification:
           * If int, the number of bins for the two dimensions
             (``nx = ny = bins``).
           * If `array_like`, the bin edges for the two dimensions
             (``x_edges = y_edges = bins``).
           * If [int, int], the number of bins in each dimension
             (``nx, ny = bins``).
           * If [`array_like`, `array_like`], the bin edges in each
             dimension (``x_edges, y_edges = bins``).
    range: array_like, shape(2,2), optional
        The edges of this histogram along each dimension. If ``bins``
        is not integral, then this parameter is ignored. If None, the
        default is ``[[x.min(), x.max()], [y.min(), y.max()]]``.
    weights: array_like
        An array of weights associated to each element :math:`(x_i,
        y_i)` pair.  Each pair of the data will contribute its
        associated weight to the bin count.
    flow : bool
        Include over/underflow.
    cons_var : bool
        If ``True``, conserve the variance rather than return the
        standard error (square root of the variance).

    Raises
    ------
    ValueError
        If `x` and `y` have different shape or either bin edge definition
        is not monotonically increasing.
    ValueError
        If the shape of `weights` is not compatible with `x` and `y`.
    TypeError
        If `x`, `y`, or `weights` are unsupported types

    See Also
    --------
    fix2d
        Used for no weight or single weight fixed bin width histograms
    var2d
        Used for no weight or single weight variable bin width histograms.

    Returns
    -------
    :py:obj:`numpy.ndarray`
        The bin counts.
    :py:obj:`numpy.ndarray`
        The standard error of each bin count, :math:`\sqrt{\sum_i w_i^2}`.

    Examples
    --------
    Gaussian distributions in 2D with automatic bin ranges:

    >>> rng = np.random.default_rng(123)
    >>> x = rng.standard_normal(size=(1000,))
    >>> y = rng.standard_normal(size=(1000,))
    >>> w = rng.uniform(0.3, 0.4, size=x.shape)
    >>> h, err = histogram2d(x, y, bins=[10, 10], weights=w)
    >>> h.shape
    (10, 10)

    """
    try:
        N = len(bins)
    except TypeError:
        N = 1

    x = np.asarray(x)
    y = np.asarray(y)
    if weights is not None:
        weights = np.asarray(weights)
    if N != 1 and N != 2:
        bins = np.asarray(bins)
        return var2d(x, y, bins, bins, weights=weights, flow=flow, cons_var=cons_var)

    elif N == 1:
        return fix2d(
            x,
            y,
            bins=bins,
            range=range,
            weights=weights,
            flow=flow,
            cons_var=cons_var,
        )

    elif N == 2:
        if isinstance(bins[0], int) and isinstance(bins[1], int):
            return fix2d(x, y, bins=bins, range=range, weights=weights, flow=flow)
        else:
            b1 = np.asarray(bins[0])
            b2 = np.asarray(bins[1])
            return var2d(x, y, b1, b2, weights=weights, flow=flow)

    else:
        raise ValueError("bins argument is not compatible")



# MIT License
#
# Copyright (c) 2021 Douglas Davis
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import codecs
import glob
import os
import re
import pathlib
import platform
import subprocess
import sys
import tempfile

import setuptools
from setuptools import setup
from setuptools.extension import Extension


CONDA_BUILD = "CONDA_BUILD" in os.environ


def has_flag(compiler, flag):
    """check if compiler has compatibility with the flag"""
    with tempfile.NamedTemporaryFile("w", suffix=".cpp") as f:
        f.write("int main (int argc, char** argv) { return 0; }")
        try:
            compiler.compile([f.name], extra_postargs=[flag])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def get_cpp_std_flag():
    compiler = setuptools.distutils.ccompiler.new_compiler()
    setuptools.distutils.sysconfig.customize_compiler(compiler)
    if has_flag(compiler, "-std=c++14"):
        return "-std=c++14"
    else:
        raise RuntimeError("C++14 (or later) compatible compiler required")


def conda_darwin_flags(flavor="inc"):
    if not CONDA_BUILD:
        return []
    pref = os.environ["PREFIX"]
    if flavor == "inc":
        return [f"-I{pref}/include"]
    elif flavor == "lib":
        return [f"-Wl,-rpath,{pref}/lib", f"-L{pref}/lib"]
    else:
        return []


def is_apple_silicon():
    return sys.platform.startswith("darwin") and "arm" in platform.processor()


def get_compile_flags(is_cpp=False):
    """get the compile flags"""
    if is_cpp:
        cpp_std = get_cpp_std_flag()
    cflags = ["-Wall", "-Wextra"]
    debug_env = os.getenv("PYGRAM11_DEBUG")
    if debug_env is None:
        cflags += ["-g0"]
    else:
        cflags += ["-g"]
    if sys.platform.startswith("darwin"):
        if is_cpp:
            cflags += ["-fvisibility=hidden", "-stdlib=libc++", cpp_std]
        if is_apple_silicon() and not CONDA_BUILD:
            cflags += ["-I/opt/homebrew/include"]
        cflags += ["-Xpreprocessor", "-fopenmp"]
        cflags += conda_darwin_flags("inc")
    else:
        if is_cpp:
            cflags += ["-fvisibility=hidden", cpp_std]
        cflags += ["-fopenmp"]
    return cflags


def get_link_flags(is_cpp=False):
    lflags = []
    if sys.platform.startswith("darwin"):
        if is_apple_silicon() and not CONDA_BUILD:
            lflags += ["-L/opt/homebrew/lib"]
        lflags += conda_darwin_flags("lib")
        lflags += ["-lomp"]
    else:
        lflags += ["-lgomp"]
    return lflags


def has_openmp():
    test_code = """
    #include <omp.h>
    #include <stdio.h>
    int main() {
      #pragma omp parallel
      printf("nthreads=%d\\n", omp_get_num_threads());
      return 0;
    }
    """
    has_omp = False
    compiler = setuptools.distutils.ccompiler.new_compiler()
    setuptools.distutils.sysconfig.customize_compiler(compiler)
    cflags = get_compile_flags()
    lflags = get_link_flags()
    tmp_dir = tempfile.mkdtemp()
    start_dir = pathlib.PosixPath.cwd()
    try:
        os.chdir(tmp_dir)
        with open("test_openmp.c", "w") as f:
            f.write(test_code)
        os.mkdir("obj")
        compiler.compile(["test_openmp.c"], output_dir="obj", extra_postargs=cflags)
        objs = glob.glob(os.path.join("obj", "*{}".format(compiler.obj_extension)))
        compiler.link_executable(objs, "test_openmp", extra_postargs=lflags)
        output = subprocess.check_output("./test_openmp")
        output = output.decode(sys.stdout.encoding or "utf-8").splitlines()
        if "nthreads=" in output[0]:
            nthreads = int(output[0].strip().split("=")[1])
            if len(output) == nthreads:
                has_omp = True
            else:
                has_omp = False
        else:
            has_omp = False
    except (
        setuptools.distutils.errors.CompileError,
        setuptools.distutils.errors.LinkError,
    ):
        has_omp = False
    finally:
        os.chdir(start_dir)

    return has_omp


def get_extensions():
    cpp_cflags = get_compile_flags(is_cpp=True)
    cpp_lflags = get_link_flags(is_cpp=True)
    return [
        Extension(
            "pygram11._backend",
            [os.path.join("src", "_backend.cpp")],
            language="c++",
            include_dirs=["extern/pybind11/include", "extern/mp11/include"],
            extra_compile_args=cpp_cflags,
            extra_link_args=cpp_lflags,
        ),
    ]


if not has_openmp():
    sys.exit(
        "\n"
        "****************************************************\n"
        "* OpenMP not available, aborting installation.     *\n"
        "* On macOS you can install `libomp` with Homebrew. *\n"
        "* On Linux check your GCC installation.            *\n"
        "****************************************************"
    )


extras_require = {
    "test": [
        "pytest",
    ],
}

extras_require["complete"] = sorted(set(sum(extras_require.values(), [])))

setup(
    use_scm_version=True,
    extras_require=extras_require,
    setup_requires=["setuptools_scm"],
    ext_modules=get_extensions(),
    zip_safe=False,
)



import struct
from typing import Sequence, Any

import numpy as np
import numpy.testing as npt
import pygram11 as pg
import pytest

RNG = np.random.default_rng(1234)
XTYPES: Sequence[Any] = (
    np.float32,
    np.float64,
    np.int32,
    np.int64,
    np.uint32,
    np.uint64,
)
if (struct.calcsize("P") * 8) == 32:
    XTYPES = (np.float32, np.int32, np.uint32)


class TestFix1D:
    XD = RNG.normal(scale=5, size=8000)
    WD = RNG.uniform(0.8, 1.2, XD.shape[0])

    def make_data(self, xtype, wtype):
        x = (self.XD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(xtype)
        w = (self.WD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(wtype)
        return x, w

    def make_data_mw(self, xtype, wtype):
        x = self.XD.astype(dtype=xtype)
        w1 = (self.WD * 0.1).astype(wtype)
        w2 = (self.WD * 1.1).astype(wtype)
        w3 = (self.WD * RNG.uniform(-0.3, 4.1, w1.shape[0])).astype(wtype)
        w = np.transpose(np.array([w1, w2, w3], dtype=wtype))
        return x, w

    @pytest.mark.parametrize("xtype", XTYPES)
    @pytest.mark.parametrize("wtype", [None, np.float32, np.float64])
    @pytest.mark.parametrize("density", [True, False])
    @pytest.mark.parametrize("flow", [True, False])
    @pytest.mark.parametrize("func", [pg.histogram, pg.fix1d])
    @pytest.mark.parametrize("cons_var", [True, False])
    @pytest.mark.OD
    def test_no_weight_and_single_weight(
        self, xtype, wtype, density, flow, func, cons_var
    ):
        if density and flow:
            assert True
            return
        x, w = self.make_data(xtype, wtype)
        n, xmin, xmax = 50, -10.1, 10.1
        res1, edge = np.histogram(
            x, weights=w, bins=n, range=(xmin, xmax), density=density
        )
        if flow:
            if w is None:
                res1[0] += np.sum(x < xmin)
                res1[-1] += np.sum(x >= xmax)
            else:
                res1[0] += w[x < xmin].sum()
                res1[-1] += w[x >= xmax].sum()

        with pg.omp_disabled():
            res0, err0 = func(
                x,
                weights=w,
                bins=n,
                range=(xmin, xmax),
                density=density,
                flow=flow,
                cons_var=cons_var,
            )
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)
        with pg.omp_forced():
            res0, err0 = func(
                x,
                weights=w,
                bins=n,
                range=(xmin, xmax),
                density=density,
                flow=flow,
                cons_var=cons_var,
            )
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)

    @pytest.mark.parametrize("xtype", XTYPES)
    @pytest.mark.parametrize("wtype", [np.float32, np.float64])
    @pytest.mark.parametrize("flow", [True, False])
    @pytest.mark.parametrize("func", [pg.histogram, pg.fix1dmw])
    @pytest.mark.OD
    def test_multiple_weights(self, xtype, wtype, flow, func):
        x, w = self.make_data_mw(xtype, wtype)
        n, xmin, xmax = 50, -10.1, 10.1
        with pg.omp_disabled():
            res0a, err0a = func(x, weights=w, bins=n, range=(xmin, xmax), flow=flow)
            for i in range(res0a.shape[1]):
                res1, edge = np.histogram(
                    x, weights=w[:, i], bins=n, range=(xmin, xmax)
                )
                if flow:
                    res1[0] += np.sum(w[:, i][x < xmin])
                    res1[-1] += np.sum(w[:, i][x >= xmax])
                npt.assert_allclose(res0a[:, i], res1, atol=0.01, rtol=1.0e-3)
        with pg.omp_forced():
            res0a, err0a = func(x, weights=w, bins=n, range=(xmin, xmax), flow=flow)
            for i in range(res0a.shape[1]):
                res1, edge = np.histogram(
                    x, weights=w[:, i], bins=n, range=(xmin, xmax)
                )
                if flow:
                    res1[0] += np.sum(w[:, i][x < xmin])
                    res1[-1] += np.sum(w[:, i][x >= xmax])
                npt.assert_allclose(res0a[:, i], res1, atol=0.01, rtol=1.0e-3)

    @pytest.mark.OD
    @pg.without_omp
    def test_cons_var_nomp(self):
        x = RNG.standard_normal(size=(500,))
        w = 0.5 * np.ones_like(x, dtype=np.float64)
        res1, var1 = pg.histogram(x, bins=12, range=(-3, 3), weights=w, cons_var=True)
        res2, edg2 = np.histogram(x, bins=12, range=(-3, 3), weights=w)
        npt.assert_allclose(res1, res2)
        npt.assert_allclose(var1, res1 * 0.5)

    @pytest.mark.OD
    @pg.with_omp
    def test_cons_var_womp(self):
        x = RNG.standard_normal(size=(500,))
        w = 0.5 * np.ones_like(x, dtype=np.float64)
        res1, var1 = pg.histogram(x, bins=12, range=(-3, 3), weights=w, cons_var=True)
        res2, edg2 = np.histogram(x, bins=12, range=(-3, 3), weights=w)
        npt.assert_allclose(res1, res2)
        npt.assert_allclose(var1, res1 * 0.5)


class TestVar1D:
    XD = RNG.normal(scale=0.8, size=2000)
    WD = RNG.uniform(0.8, 1.2, XD.shape[0])
    E1 = np.array(
        [-3.1, -2.2, -2.1, -1.975, -1.1, -0.9, -0.5, 0.05, 0.1, 1.5, 2.9, 3.1]
    )
    E2 = np.array(
        [-3.1, -2.1, -1.9, -1.451, -1.1, -0.5, -0.1, 0.09, 0.4, 2.2, 2.8, 3.1]
    )

    def make_data(self, xtype, wtype):
        x = (self.XD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(xtype)
        w = (self.WD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(wtype)
        return x, w

    def make_data_mw(self, xtype, wtype):
        x = self.XD.astype(dtype=xtype)
        w1 = (self.WD * 0.1).astype(wtype)
        w2 = (self.WD * 1.1).astype(wtype)
        w3 = (self.WD * RNG.uniform(-0.3, 3.1, w1.shape[0])).astype(wtype)
        w = np.transpose(np.array([w1, w2, w3], dtype=wtype))
        return x, w

    @pytest.mark.parametrize("xtype", XTYPES)
    @pytest.mark.parametrize("wtype", [None, np.float32, np.float64])
    @pytest.mark.parametrize("bins", [E1, E2])
    @pytest.mark.parametrize("density", [True, False])
    @pytest.mark.parametrize("flow", [True, False])
    @pytest.mark.parametrize("func", [pg.histogram, pg.var1d])
    @pytest.mark.OD
    def test_no_weight_and_single_weight(self, xtype, wtype, bins, density, flow, func):
        if density and flow:
            assert True
            return
        x, w = self.make_data(xtype, wtype)
        res1, edge = np.histogram(x, weights=w, bins=bins, density=density)
        xmin, xmax = bins[0], bins[-1]
        if flow:
            if w is None:
                res1[0] += np.sum(x < xmin)
                res1[-1] += np.sum(x >= xmax)
            else:
                res1[0] += w[x < xmin].sum()
                res1[-1] += w[x >= xmax].sum()

        with pg.omp_forced():
            res0, err0 = func(x, weights=w, bins=bins, density=density, flow=flow)
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)
        with pg.omp_disabled():
            res0, err0 = func(x, weights=w, bins=bins, density=density, flow=flow)
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)

    @pytest.mark.parametrize("xtype", XTYPES)
    @pytest.mark.parametrize("wtype", [np.float32, np.float64])
    @pytest.mark.parametrize("bins", [E1, E2])
    @pytest.mark.parametrize("flow", [True, False])
    @pytest.mark.parametrize("func", [pg.var1dmw, pg.histogram])
    @pytest.mark.OD
    def test_multiple_weights(self, xtype, wtype, bins, flow, func):
        x, w = self.make_data_mw(xtype, wtype)
        xmin, xmax = bins[0], bins[-1]
        res0, err0 = func(x, weights=w, bins=bins, flow=flow)
        for i in range(res0.shape[1]):
            res1, edge = np.histogram(x, weights=w[:, i], bins=bins)
            if flow:
                res1[0] += np.sum(w[:, i][x < xmin])
                res1[-1] += np.sum(w[:, i][x >= xmax])
            npt.assert_allclose(res0[:, i], res1, atol=0.01, rtol=1.0e-3)


class TestFix2D:
    XD = RNG.normal(scale=5.0, size=8000)
    YD = RNG.normal(scale=6.0, size=8000)
    WD = RNG.uniform(0.8, 1.2, XD.shape[0])

    def make_data(self, xtype, ytype, wtype):
        x = (self.XD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(xtype)
        y = (self.YD * RNG.uniform(0.9, 1.1, self.YD.shape[0])).astype(ytype)
        w = (self.WD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(wtype)
        return x, y, w

    @pytest.mark.parametrize("xtype", XTYPES)
    @pytest.mark.parametrize("ytype", XTYPES)
    @pytest.mark.parametrize("wtype", [None, np.float64, np.float32])
    @pytest.mark.parametrize("flow", [False])
    @pytest.mark.parametrize("func", [pg.histogram2d, pg.fix2d])
    @pytest.mark.TD
    def test_no_weight_and_single_weight(self, xtype, ytype, wtype, flow, func):
        x, y, w = self.make_data(xtype, ytype, wtype)
        nbx, xmin, xmax = 25, -10.1, 10.1
        nby, ymin, ymax = 15, -10.1, 10.1
        res1, ex, ey = np.histogram2d(
            x, y, bins=[nbx, nby], range=((xmin, xmax), (ymin, ymax)), weights=w
        )

        with pg.omp_disabled():
            res0, err0 = func(
                x,
                y,
                bins=[nbx, nby],
                range=((xmin, xmax), (ymin, ymax)),
                weights=w,
                flow=flow,
            )
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)
        with pg.omp_forced():
            res0, err0 = func(
                x,
                y,
                bins=[nbx, nby],
                range=((xmin, xmax), (ymin, ymax)),
                weights=w,
                flow=flow,
            )
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)


class TestVar2D:
    XD = RNG.normal(scale=0.9, size=2000)
    YD = RNG.normal(scale=0.9, size=2000)
    WD = RNG.uniform(0.8, 1.2, XD.shape[0])
    E1 = np.array(
        [-3.1, -2.2, -2.1, -1.975, -1.1, -0.9, -0.5, 0.05, 0.1, 1.5, 2.9, 3.1]
    )
    E2 = np.array(
        [-3.1, -2.1, -1.9, -1.451, -1.1, -0.5, -0.1, 0.09, 0.4, 2.2, 2.8, 3.1]
    )

    def make_data(self, xtype, ytype, wtype):
        x = (self.XD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(xtype)
        y = (self.YD * RNG.uniform(0.9, 1.1, self.YD.shape[0])).astype(ytype)
        w = (self.WD * RNG.uniform(0.9, 1.1, self.XD.shape[0])).astype(wtype)
        return x, y, w

    @pytest.mark.parametrize("xtype", XTYPES)
    @pytest.mark.parametrize("ytype", XTYPES)
    @pytest.mark.parametrize("wtype", [None, np.float64, np.float32])
    @pytest.mark.parametrize("xbins", [E1, E2])
    @pytest.mark.parametrize("ybins", [E1, E2])
    @pytest.mark.parametrize("flow", [False])
    @pytest.mark.parametrize("func", [pg.var2d, pg.histogram2d])
    @pytest.mark.TD
    def test_no_weight_and_single_weight(
        self, xtype, ytype, wtype, xbins, ybins, flow, func
    ):
        x, y, w = self.make_data(xtype, ytype, wtype)
        res1, edgex, edgey = np.histogram2d(x, y, bins=[xbins, ybins], weights=w)

        with pg.omp_disabled():
            if func == pg.histogram2d:
                res0, err0 = func(x, y, bins=[xbins, ybins], weights=w, flow=flow)
            elif func == pg.var2d:
                res0, err0 = func(x, y, xbins, ybins, weights=w, flow=flow)
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)
        with pg.omp_forced():
            if func == pg.histogram2d:
                res0, err0 = func(x, y, bins=[xbins, ybins], weights=w, flow=flow)
            elif func == pg.var2d:
                res0, err0 = func(x, y, xbins, ybins, weights=w, flow=flow)
            npt.assert_allclose(res0, res1, atol=0.01, rtol=1.0e-3)


BAD_TYPES = (np.int8, np.int16, np.uint8, np.uint16, np.float16)


class TestExceptions:
    X = RNG.uniform(0, 100, 50)
    W1 = RNG.uniform(0.5, 1.5, 50).astype(np.float64)
    W2 = np.abs(RNG.standard_normal(size=(50, 3)))

    @pytest.mark.parametrize("xtype", BAD_TYPES)
    @pytest.mark.misci
    def test_f1d(self, xtype):
        x = self.X.astype(xtype)
        w1 = self.W1
        w2 = self.W2
        # bad type
        with pytest.raises(TypeError):
            pg.fix1d(x)

        # bad type
        with pytest.raises(TypeError):
            pg.fix1d(x, weights=w1)

        # bad shape
        with pytest.raises(ValueError):
            pg.fix1d(x, weights=w2)

        # bad type
        with pytest.raises(TypeError):
            pg.fix1dmw(x, weights=w2)

    @pytest.mark.parametrize("xtype", BAD_TYPES)
    @pytest.mark.misci
    def test_v1d(self, xtype):
        x = self.X.astype(xtype)
        # bad range
        with pytest.raises(ValueError):
            pg.histogram(x.astype(np.float32), bins=[1.0, 2.0, 3.0, 5.0], range=(-1, 1))


class TestConvenience:
    @pytest.mark.misci
    def test_bin_centers(self):
        edges = [1, 2, 3, 4, 5, 6]
        c = pg.bin_centers(edges)
        c2 = np.array([1.5, 2.5, 3.5, 4.5, 5.5])
        npt.assert_allclose(c, c2)
        c = pg.bin_centers(5, (-3, 2))
        c2 = np.array([-2.5, -1.5, -0.5, 0.5, 1.5])
        npt.assert_allclose(c, c2)
        c = pg.bin_centers([1, 3, 4, 8, 9, 9.2])
        c2 = np.array([2.0, 3.5, 6.0, 8.5, 9.1])
        npt.assert_allclose(c, c2)

    @pytest.mark.misci
    def test_bin_edges(self):
        edges = pg.bin_edges(8, (-4, 4))
        e2 = np.array([-4.0, -3.0, -2.0, -1.0, 0.0, 1.0, 2.0, 3.0, 4.0])
        npt.assert_allclose(edges, e2)
        edges = pg.bin_edges(3, (0, 3))
        e2 = np.array([0.0, 1.0, 2.0, 3.0])
        npt.assert_allclose(edges, e2)



"""pygram11 configuration."""


from typing import Dict, Iterator


config: Dict[str, int] = {
    "thresholds.fix1d": 10_000,
    "thresholds.fix1dmw": 10_000,
    "thresholds.fix2d": 10_000,
    "thresholds.var1d": 5_000,
    "thresholds.var1dmw": 5_000,
    "thresholds.var2d": 5_000,
}


def get(key: str) -> int:
    """Retrieve a configuration value given a key.

    Parameters
    ----------
    key : str
        Desired configuration parameter key.

    Returns
    -------
    int
        Value of the key.

    """
    global config
    return config[key]


def set(key: str, value: int) -> None:
    """Set a configuration key's value.

    Parameters
    ----------
    key : str
        Desired configuration parameter key to modify.
    value : int
        New value for the given key.

    """
    global config
    if key not in config:
        raise ValueError(f"{key} not in config")
    config[key] = value


def threshold_keys() -> Iterator[str]:
    """All available keys in the configuration dictionary."""
    global config
    for k in config.keys():
        yield k
